! FORTH Block conversion swiss army knife 
! Nils M Holm, 2021
! Public domain / CC0 License

! Usage: block [-asu0] [-f N] [-i N] [-r N N] [-t S] [-x N N]
!              [infile [outfile]]
!
! -0     add empty block 0 when coverting text to block
! -a     append blocks or text to outfile
! -b     copy binary data to outfile, fill to block boundary
! -f n   output N empty blocks (use -f n - outfile)
! -i n   extract system image, N = bytes as reported by SAVE-SYSTEM
! -j n   extract secondary system image, see above
! -r n r reserve N additional blocks, then round up to multiple of R
! -s     strip block headings and collapse consecutive blank lines
! -t s   Use S as title comment for each screen
! -u     unblock: convert blocks to text file
! -x n m extract blocks N through M
!
! When infile and outfile are given, convert infile to outfile.
! When only infile is given, convert infile.fth to outfile.blk.
! When no file is given, convert sysin to sysout.
! An infile of "-" is sysin, an outfile of "-" is sysout.
!
! File name length is limited to 14 characters. Enough for CP/M
! and local file names on DOS.

module block(t3x);

object	t[t3x];

const	BUFLEN = 1024;

var	Buffer::BUFLEN;
var	Infile;
var	Bp, Nchars;

var	Screen::1024;
var	Blanks::1024;
var	Scr, Lno;
var	Title::65;

var	Outfile;
var	Block0;

str_length(s) return t.memscan(s, 0, 32767);

writes(s) t.write(Outfile, s, str_length(s));

nl() do var b::3;
	t.newline(b);
	t.write(Outfile, b, str_length(b));
end

var	ntoa_buf::32;

ntoa(v, w) do var i;
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0';
                i := i - 1;
                v := v / 10;
		w := w-1;
        end
	while (w > 0) do
                ntoa_buf::i := '\s';
                i := i - 1;
		w := w-1;
	end
        return @ntoa_buf::(i+1);
end

aton(s) do var v, i;
	v := 0;
	i := 0;
	while ('0' <= s::i /\ s::i <= '9') do
		v := v * 10 + s::i - '0';
		i := i+1;
	end
	return v;
end

error(m, s) do var b::3;
	t.write(T3X.SYSERR, "block: error: ", 14);
	t.write(T3X.SYSERR, m, str_length(m));
	if (s) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, s, str_length(s));
	end
	t.newline(b);
	t.write(T3X.SYSERR, b, str_length(b));
	halt 1;
end

rdchar() do
	if (Bp >= Nchars) do
		Nchars := t.read(Infile, Buffer, BUFLEN);
		if (Nchars < 1) return %1;
		Bp := 0;
	end
	Bp := Bp+1;
	return Buffer::(Bp-1);
end

var	Line::65;

readline() do var i, c, tab;
	tab := 1;
	for (i=0, 65) do
		c := rdchar();
		! skip leading TAB
		if (tab /\ i = 0 /\ c = '\t') do
			tab := 0;
			i := i-1;
			loop;
		end
		if (c < 0 /\ i = 0) return %1;
		if (c = '\r') loop;
		if (c = '\n') leave;
		Line::i := c;
	end
	Line::i := 0;
	if (c \= '\n') error("line too long", Line);
	return Line;
end

insert_title(s) do var k;
	s:: 0 := '(';
	k := str_length(Title);
	t.memcopy(@s::2, Title, k);
	s::(k+3) := ')';
end

var	Aux::1024;

! Create new screen, coping I lines
! over from the previous screen

newscreen(i) do var k;
	i := i+1;
	if (i > 1) t.memcopy(Aux, @Screen::(i*64), 1024-i*64);
	Lno := 2;
	t.memfill(Screen, '\s', 1024);
	ie (Title::0) do
		insert_title(Screen);
	end
	else do
		Screen:: 0 := '(';
		t.memcopy(@Screen::2, "Screen ", 9);
		t.memcopy(@Screen::9, ntoa(Scr, 4), 4);
		Screen::14 := ')';
	end
	if (i > 1) do
		t.memcopy(@Screen::128, Aux, 1024-i*64);
		Lno := Lno + (1024-i*64) / 64;
	end
	Scr := Scr+1;
end

! Copy line S to record D and fill
copy(d, s) do
	t.memfill(d, '\s', 64);
	t.memcopy(d, s, str_length(s));
end

isblank(s) do var i;
	for (i=0, 64)
		if (s::i \= '\s')
			return 0;
	return %1;
end

! line N empty?
empty(n) return isblank(@Screen::(64*n));

! Write screen (block). When the last line would not
! be empty, postpone output of the last paragraph to
! the next screen. When that screen would also overflow,
! split the paragraph.

writescreen() do var i;
	for (i=15, 1, %1)
		if (empty(i))
			leave;
	if (i = 1) do
		t.write(Outfile, Screen, 1024);
		return 15;
	end
	t.write(Outfile, Screen, (i+1)*64);
	if (i \= 15) t.write(Outfile, Blanks, 1024-(i+1)*64);
	return i;
end

! Output empty screens/blocks with optional title.
fill(n) do var i;
	t.memfill(Screen, '\s', 1024);
	if (Title::0) insert_title(Screen);
	for (i=0, n) t.write(Outfile, Screen, 1024);
end

! Convert text file to blocks.
block(rn, rr) do var i, k;
	k := 0;
	ie (Block0) do
		newscreen(0);
		i := writescreen();
		k := k+1;
	end
	else do
		i := 0;
		Scr := 1;
	end
	newscreen(i);
	while (readline() \= %1) do
		copy(@Screen::(Lno*64), Line);
		Lno := Lno + 1;
		if (Lno = 16) do
			i := writescreen();
			k := k+1;
			newscreen(i);
		end
	end
	writescreen();
	k := k+1;
	if (rn \= %1) do
		fill(rn);
		k := k+rn;
		if (rr \= 1) fill(rr - k mod rr);
	end
end

! Output empty binary blocks
binfill(n) do var i;
	t.memfill(Screen, 0, 1024);
	for (i=0, n) t.write(Outfile, Screen, 1024);
end

! Output binary data to blocks.
binblock(rn, rr) do var k, n;
	k := 0;
	t.memfill(Screen, 0, 1024);
	n := t.read(Infile, Screen, 1024);
	while (n > 0) do
		if (t.write(Outfile, Screen, 1024) \= 1024)
			error("file write error", 0);
		k := k+1;
		t.memfill(Screen, 0, 1024);
		n := t.read(Infile, Screen, 1024);
	end
	if (rn \= %1) do
		binfill(rn);
		k := k+rn;
		if (rr \= 1) binfill(rr - k mod rr);
	end
end

! Right-trim S.
trim(s) do var i;
	for (i=63, %1, %1) do
		if (s::i \= '\s') return i+1;
	end
	return 0;
end

var	Blank;

writeline(s, sq) do var nb;
	nb := isblank(s);
	if (sq /\ Blank /\ nb) return;
	Blank := nb;
	t.write(Outfile, s, trim(s));
	nl();
end

! Convert blocks to text file.
unblock(x1, xn, sq) do var i, j, k;
	j := 0;
	if (x1 \= %1) do
		for (j=0, x1) do
			k := t.read(Infile, Screen, 1024);
			if (k \= 1024)
				error("file read error", 0);
		end
	end
	Blank := 0;
	while (1) do
		if (xn \= %1 /\ j > xn) leave;
		k := t.read(Infile, Screen, 1024);
		if (k < 1) do
			if (xn = %1) leave;
			error("file read error", 0);
		end
		if (k \= 1024) error("strange file size", 0);
		for (i=sq-> 2: 0, 16) do
			writeline(@Screen::(i*64), sq);
		end
		j := j+1;
	end
end

! extract specific blocks
extract_image(n, off) do var i;
	for (i=0, off) do
		if (t.read(Infile, Screen, 1024) \= 1024)
			error("file read error", 0);
	end
	while (n >= 1024) do
		if (t.read(Infile, Screen, 1024) \= 1024)
			error("file read error", 0);
		if (t.write(Outfile, Screen, 1024) \= 1024)
			error("file write error", 0);
		n := n - 1024;
	end
	if (n) do
		if (t.read(Infile, Screen, n) \= n)
			error("file read error", 0);
		if (t.write(Outfile, Screen, n) \= n)
			error("file write error", 0);
	end
end

init() do
	Scr := 0;
	Title::0 := 0;
	Block0 := 0;
	t.memfill(Blanks, '\s', 1024);
end

open(s, mode) do
	if (t.memcomp(s, "-", 2) = 0)
		return mode = T3X.OREAD-> T3X.SYSIN:
					T3X.SYSOUT;
	return t.open(s, mode);
end

do
	var	in::14, out::14, arg::14, ki, ko;
	var	i, na, u, k, mode, fb, ex1, ex2, im, off, sq, rn, rr, bin;

	init();
	na := 1;
	u := 0;
	fb := 0;
	ex1 := %1;
	rn := %1;
	rr := 1;
	im := 0;
	bin := 0;
	sq := 0;
	mode := T3X.OWRITE;
	while (1) do
		i := t.getarg(na, arg, 14);
		ie (t.memcomp(arg, "-a", 3) = 0) do
			mode := T3X.OAPPND;
			na := na + 1;
		end
		else ie (t.memcomp(arg, "-b", 3) = 0) do
			bin := 1;
			na := na + 1;
		end
		else ie (t.memcomp(arg, "-f", 3) = 0) do
			k := t.getarg(na+1, arg, 60);
			fb := aton(arg);
			na := na + 2;
		end
		else ie (t.memcomp(arg, "-i", 3) = 0) do
			k := t.getarg(na+1, arg, 60);
			off := 1;
			im := aton(arg);
			na := na + 2;
		end
		else ie (t.memcomp(arg, "-j", 3) = 0) do
			k := t.getarg(na+1, arg, 60);
			off := 33;
			im := aton(arg);
			na := na + 2;
		end
		else ie (t.memcomp(arg, "-x", 3) = 0) do
			k := t.getarg(na+1, arg, 60);
			ex1 := aton(arg);
			k := t.getarg(na+2, arg, 60);
			ex2 := aton(arg);
			na := na + 3;
		end
		else ie (t.memcomp(arg, "-r", 3) = 0) do
			k := t.getarg(na+1, arg, 60);
			rn := aton(arg);
			k := t.getarg(na+2, arg, 60);
			rr := aton(arg);
			na := na + 3;
		end
		else ie (t.memcomp(arg, "-u", 3) = 0) do
			u := %1;
			na := na + 1;
		end
		else ie (t.memcomp(arg, "-s", 3) = 0) do
			sq := %1;
			na := na + 1;
		end
		else ie (t.memcomp(arg, "-0", 3) = 0) do
			Block0 := 1;
			na := na + 1;
		end
		else ie (t.memcomp(arg, "-t", 3) = 0) do
			k := t.getarg(na+1, Title, 60);
			for (i=0, k)
				if (Title::i = '_')
					Title::i := '\s';
			na := na+2;
		end
		else do
			leave;
		end
	end
	ki := t.getarg(na, in, 13);
	ko := t.getarg(na+1, out, 13);
	ie (u = 0 /\ ki > 0) do
		if (ko < 1) do
			if (ki > 8) ki := 8;
			ko := ki;
			t.memcopy(out, in, ki+1);
			t.memcopy(@in::ki, ".fth", 5);
			t.memcopy(@out::ko, ".blk", 5);
		end
		Infile := open(in, T3X.OREAD);
		if (Infile < 0) error("no such file", in);
		Outfile := open(out, mode);
		if (Outfile < 0) error("cannot create file", out);
	end
	else ie (u \= 0 /\ ki > 0) do
		Infile := open(in, T3X.OREAD);
		if (Infile < 0) error("no such file", in);
		ie (ko > 0) do
			Outfile := open(out, mode);
			if (Outfile < 0) error("cannot create file", out);
		end
		else do
			Outfile := T3X.SYSOUT;
		end
	end
	else do
		Infile := T3X.SYSIN;
		Outfile := T3X.SYSOUT;
	end
	ie (fb)
		fill(fb);
	else ie (ex1 \= %1)
		unblock(ex1, ex2, sq);
	else ie (im)
		extract_image(im, off);
	else ie (u)
		unblock(%1, %1, sq);
	else ie (bin)
		binblock(rn, rr);
	else
		block(rn, rr);
end
