! Colorize/HTMLize the T3XFORTH source code
! Nils M Holm, 2001,2004,2009,2021
! Public domain / CC0 license

! Highly specific to T3XFORTH and probably full of bugs.

! Reads the T3XFORTH image to find out which definition
! is where. Assumes that
! 1) definitions defined locally are local;
! 2) definitions contained in the image are in T3XFORTH.FTH;
! 3) definition not contained anywhere are in T3XFORTH.S86.

module colorize(t3x);

object	t[t3x];

const	BUFLEN	= 1024;
const	MAXWORD	= 128;

var	Infile, Outfile;
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Inp, Inlim, Outp;
var	Rejected;
var	Line;
var	C_numlit,
	C_strlit,
	C_comment,
	C_def,
	C_text,
	C_bg,
	C_href;

var	Mem::32760, Mtop;
var	Forth;

var	Names::4096, Ntop;

mword(a) return Mem::(a+1) << 8 | Mem::a;

lfa(a) return  a+6;
pfa(a) return a+10;

flush() do
	if (Outp > 0) t.write(Outfile, Outbuf, Outp);
	Outp := 0;
end

fini() do
	flush();
	t.close(Infile);
	t.close(Outfile);
end

str_length(s) return t.memscan(s, 0, 32767);

str_equal(a, b) return t.memcomp(a, b, str_length(a)+1) = 0;

str_copy(a, b) t.memcopy(a, b, str_length(b)+1);

var	ntoa_buf::32;
ntoa(v, w) do var i;
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0';
                i := i - 1;
                v := v / 10;
		w := w-1;
        end
	while (w > 0) do
                ntoa_buf::i := '\s';
                i := i - 1;
		w := w-1;
	end
        return @ntoa_buf::(i+1);
end

error(s) do
	t.write(T3X.SYSERR, "Colorize: error: ", 17);
	t.write(T3X.SYSERR, s, str_length(s));
	t.newline(s);
	t.write(T3X.SYSERR, s, str_length(s));
	flush();
	halt 1;
end

rdch() do var c;
	if (Inp >= Inlim) do
		Inlim := t.read(Infile, Inbuf, BUFLEN);
		if (Inlim < 1) return %1;
		Inp := 0;
	end
	Inp := Inp + 1;
	return Inbuf::(Inp-1);
end

wrch(c) do
	if (Outp >= BUFLEN) flush();
	Outbuf::Outp := c;
	Outp := Outp + 1;
end

writes(s) do var i, k;
	k := str_length(s);
	for (i=0, k) wrch(s::i);
end

coreword(s) do var i, a, k, n;
	k := str_length(s);
	n := k > 5-> 5: k;
	i := Forth;
	while (i \= 0) do
		a := i - 256;
		if (	Mem::a & 0x1F = k /\
			t.memcomp(@Mem::(a+1), s, n) = 0
		)
			return %1;
		i := mword(lfa(a));
	end
	return 0;
end

init() do
	Rejected := 0;
	Inp := 0;
	Inlim := 0;
	Outp := 0;
	Line := 1;
	Ntop := 0;
	C_text :=     "t";
	C_numlit :=   "n";
	C_strlit :=   "s";
	C_comment :=  "c";
	C_def :=      "d";
end

initwords() do var fd, i;
	fd := t.open("t3xforth.com", T3X.OREAD);
	if (fd >= 0) do
		Mtop := t.read(fd, Mem, 32760);
		t.close(fd);
		for (i=0, Mtop)
			if (t.memcomp(@Mem::i,
				packed[7,'(','F','O','R','T'], 6) = 0
			)
				leave;
		ie (i \= Mtop)
			Forth := mword(pfa(i));
		else
			Forth := 0;
	end
end

readword(s, dlm) do var c, i;
	if (Rejected) do
		wrch(Rejected);
		Rejected := 0;
	end
	c := rdch();
	ie (dlm) do
		while (c = dlm) do
			wrch(c);
			c := rdch();
		end
	end
	else do
		while (c = ' ' \/ c = '\n' \/ c = '\r') do
			wrch(c);
			c := rdch();
		end
	end
	i := 0;
	ie (dlm) do
		while (c \= %1 /\ c \= dlm) do
			if (i > MAXWORD) error("word too long");
			s::i := c;
			i := i+1;
			c := rdch();
		end
	end
	else do
		while (c \= %1 /\ c \= ' ' /\ c \= '\n' /\ c \= '\r') do
			if (i > MAXWORD) error("word too long");
			s::i := c;
			i := i+1;
			c := rdch();
		end
	end
	Rejected := c;
	s::i := 0;
end

htmlize(s) do
	var	i;

	i := 0;
	while (s::i) do
		ie (s::i = '<') writes("&lt;");
		else ie (s::i = '>') writes("&gt;");
		else ie (s::i = '&') writes("&amp;");
		else ie (s::i = '-') writes("&minus;");
		else wrch(s::i);
		i := i+1;
	end
	return s;
end

linkify(s) do
	var	i;

	i := 0;
	while (s::i) do
		ie (s::i = '?') wrch('Q');
		else ie (s::i = '"') wrch('~');
		else ie (s::i = '<') writes("&lt;");
		else ie (s::i = '>') writes("&gt;");
		else ie (s::i = '&') writes("&amp;");
		else wrch(s::i);
		i := i+1;
	end
	return s;
end

iscode(s) do var codewords, i;
	codewords := [ "!", "(+LOOP)", "(/LOOP)", "(0BRANCH)",
			"(BRANCH)", "(COLON)", "(CONST)", "(DO)",
			"(DOES>)", "(LIT)", "(LOOP)", "(VAR)", "0<",
			">R", "?SAME", "@", "AND", "BYE", "C!",
			"C@", "COLD", "DROP", "DUP", "EXECUTE",
			"EXIT", "I", "LSHIFT", "OR", "OVER", "R>",
			"R@", "READ-BLOCK", "RP!", "RP@", "RSHIFT",
			"RX?", "SP!", "SP@", "SWAP", "TX!", "UM+",
			"WRITE-BLOCK", "XOR", 0 ];
	i := 0;
	while (codewords[i]) do
		if (str_equal(s, codewords[i])) return %1;
		i := i+1;
	end
	return 0;
end

addname(s) do var k;
	k := str_length(s);
	if (Ntop + 6 >= 4096) return;
	Names::Ntop := k;
	k := k > 5-> 5: k;
	t.memcopy(@Names::(Ntop+1), s, k);
	Ntop := Ntop + k+1;
end

findname(s) do var k, n, i, j;
	k := str_length(s);
	n := k > 5-> 5: k;
	i := 0;
	while (i < Ntop) do
		j := Names::i;
		if (j = k /\ t.memcomp(@Names::(i+1), s, n) = 0)
			return %1;
		j := j>5-> 5: j;
		i := i+j+1;
	end
	return 0;
end

link(s) do
	writes("<A href=\q");
	ie (iscode(s)) writes("t3xfprim.html");
	else ie (findname(s)) ; ! write nothing
	else ie (coreword(s)) writes("t3xforth.html");
	else ; ! write nothing
	writes("#");
	linkify(s);
	writes("\q>");
	htmlize(s);
	writes("</A>");
end

name(s) do
	writes("<A name=\q");
	linkify(s);
	writes("\q>");
	htmlize(s);
	writes("</A>");
end

setcolor(c) do var buf::100;
	writes("<SPAN class=\q");
	writes(c);
	writes("\q>");
end

endcolor() writes("</SPAN>");

literal(s) do
	setcolor(C_numlit);
	htmlize(s);
	endcolor();
end

strlit(s) do
	setcolor(C_strlit);
	htmlize(s);
	endcolor();
end

comment(dlm) do var s::MAXWORD+1, i, c;
	if (Rejected = '\n') return;
	setcolor(C_comment);
	i := 1;
	s::0 := '\s';
	c := rdch();
	while (c \= %1 /\ c \= dlm) do
		s::i := c;
		i := i+1;
		c := rdch();
	end
	s::i := 0;
	htmlize(s);
	endcolor();
end

code(s) link(s);

define(s) do
	setcolor(C_def);
	writes("<B>");
	name(s);
	addname(s);
	writes("</B>");
	endcolor();
end

numeric(s) do var i;
	i := s::0 = '-'-> 1: 0;
	while (s::i) do
		if (s::i < '0' \/ s::i > '9') return 0;
		i := i+1;
	end
	return -1;
end

process(defcore) do
	var	word::MAXWORD+1;
	var	defn, last, state;
	var	i;

	defn := 0;
	last := 0;
	state := 0;
	while (1) do
		readword(word, 0);
		if (word::0 = 0) leave;
		if (	\defn /\
			str_equal(word, "\\")
		) do
			code(word);
			comment('\n');
			Rejected := '\n';
			defn := 0;
			last := 0;
			loop;
		end
		if (	\defn /\
			str_equal(word, "(")
		) do
			code(word);
			comment(')');
			Rejected := 0;
			literal(")");
			defn := 0;
			last := 0;
			loop;
		end
		if (	\defn /\
			(str_equal(word, "ABORT\q") \/
			 str_equal(word, ".\q") \/
			 str_equal(word, "$\q"))
		) do
			code(word);
			readword(word, '"');
			strlit(word);
			Rejected := 0;
			literal("\q");
			defn := 0;
			last := 0;
			loop;
		end
		if (\defn /\ str_equal(word, "DEFER")) do
			code(word);
			readword(word, 0);
			writes("<A name=\q");
			linkify(word);
			writes("\q></A>");
			writes("<A href=\q#(");
			linkify(word);
			writes(")\q>");
			htmlize(word);
			writes("</A>");
			defn := 0;
			last := 0;
			loop;
		end
		last := defn;
		defn := last-> 0:
			state-> 0:
			str_equal(word, ":") \/
			str_equal(word, "VARIABLE") \/
			str_equal(word, "CONSTANT") \/
			str_equal(word, "ARRAY") \/
			str_equal(word, "VOCABULARY") -> 1: 0;
		if (str_equal(word, ":")) state := 1;
		if (str_equal(word, ";")) state := 0;
		if (\last /\ numeric(word)) do
			ie (findname(word) \/ coreword(word) /\ \defcore)
				code(word);
			else
				literal(word);
			defn := 0;
			loop;
		end
		if (\last /\
			(str_equal(word, "[CHAR]") \/
			 str_equal(word, "[HEX]"))
		) do
			code(word);
			readword(word, 0);
			literal(word);
			defn := 0;
			loop;
		end
		if (\last /\
			(str_equal(word, "COMPILE") \/
			 str_equal(word, "[COMPILE]"))
		) do
			code(word);
			readword(word, 0);
			code(word);
			defn := 0;
			loop;
		end
		ie (last)
			define(word);
		else
			code(word);
	end
end

process2() do
	var	word::MAXWORD+1;
	var	state, last;

	state := 0;
	last := 0;
	while (1) do
		readword(word, 0);
		if (word::0 = 0) leave;
		last := state;
		state := last-> 0:
			str_equal(word, "CODE") -> 1: 0;
		ie (last)
			define(word);
		else
			htmlize(word);
	end
end

head() do var b::128;
	writes("<!DOCTYPE HTML PUBLIC");
	writes(" \q-//W3C//DTD HTML 4.0 Transitional//EN\q>\n");
	writes("<HTML>\n");
	writes("<HEAD>\n");
	writes("<TITLE>Hyperlinked T3XFORTH Source Code</TITLE>\n");
	writes("<STYLE type=\qtext/css\q>\n");
	writes("PRE.forth { background:  #000000;\n"); 
	writes("            font-size:   larger;\n");
	writes("            color:       #c0c0c0;  }\n");
	writes("SPAN.t    { font-style:  normal;   } /* text */\n");
	writes("SPAN.n    { color:       #e0e080;  } /* literal */\n");
	writes("SPAN.s    { color:       #ffa080;  } /* string */\n");
	writes("SPAN.c    { font-style:  italic;    /* comment */\n");
	writes("            color:       #c0c0c0;  }\n");
	writes("SPAN.d    { font-weight: bold;      /* definition */\n");
	writes("            color:       #a0ffa0;  }\n");
	writes("A         { text-decoration: none; } /* thread */\n");
	writes("A:link    { color:       #90c0ff;  } /* thread */\n");
	writes("A:visited { color:       #90c0ff;  } /* thread */\n");
	writes("A:hover   { color:       #000000;\n");
	writes("            background:  #90c0ff;  }\n");
	writes("</STYLE>\n");
	writes("</HEAD>\n");
	writes("<BODY>\n");
	writes("<PRE class=forth>\n");
end

tail() do
	writes("</PRE>\n");
	writes("</BODY>\n");
	writes("</HTML>\n");
end

do var arg::14, out::15, i, j, k;
	initwords();
	Infile := t.open("t3xforth.fth", T3X.OREAD);
	Outfile := t.open("t3xforth.html", T3X.OWRITE);
	init();
	head();
	process(%1);
	tail();
	fini();

	Infile := t.open("t3xforth.s86", T3X.OREAD);
	Outfile := t.open("t3xfprim.html", T3X.OWRITE);
	init();
	head();
	process2();
	tail();
	fini();

	i := 1;
	while (1) do
		k := t.getarg(i, arg, 14);
		if (k < 1) leave;
		for (j=0, k) if (arg::j = '.') leave;
		if (arg::j \= '.') t.memcopy(@arg::j, ".fth", 5);
		Infile := t.open(arg, T3X.OREAD);
		if (Infile < 0) error(arg);
		t.memcopy(out, arg, k+1);
		for (j=0, k) if (out::j = '.') leave;
		t.memcopy(@out::j, ".html", 6);
		Outfile := t.open(out, T3X.OWRITE);
		init();
		head();
		process(0);
		tail();
		fini();
		i := i+1;
	end
end

