/*
 * An ad-hoc T3XFORTH VM Emulator for Unix.
 * Nils M Holm, 2021
 * In the public domain / under the CC0 license
 */

/*
 * This emulator can run a T3XFORTH image under Unix as
 * long as it does not use any CODE words. It does not
 * emulate an 8086 processor, but just extracts the PFAs
 * of T3XFORTH primitives, intercepts dispatch to them,
 * and emulates them in C functions.
 *
 * When RX? finds the receiver empty, it will sleep for
 * 50 milliseconds so your modern CPU will not start to
 * glow while the system is idle. Crude, but works for me.
 *
 * When you happen to crash T3XFORTH, press the interrupt
 * key. You can then (b)reak program execution and
 * (hopefully) get back to the prompt, or (r)eboot the
 * system, which will reload a clean image, or (q)uit
 * from the VM. Any other key will continue execution.
 * Defining RAWMODE below will disable the interrupt key
 * and allow T3XFORTH to process any key code.
 *
 * Usage: compile a T3XFORTH image, compile this file,
 * and run it.
 */

/* #define RAWMODE */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>
#include <signal.h>

#define byte	unsigned char
#define uint	unsigned int

#define nfa(x)	((x)-10)
#define lfa(x)	((x)+6)
#define cfa(x)	((x)+8)
#define pfa(x)	((x)+10)

byte	M[65536];

uint	Barrier;
void	(*Prim[1280])();

uint	Sp, Bp, Si, Ax, Bx, Dx;
int	Bdev;

struct prim_ {
	char	*name;
	void	(*fn)(void);
};

#define PRIM struct prim_

void quit(char *s, char *a) {
	write(2, "TFVM: ", 6);
	write(2, s, strlen(s));
	if (a) {
		write(2, ": ", 2);
		write(2, a, strlen(a));
	}
	write(2, "\n", 1);
	exit(1);
}

uint word(uint a) {
	return (M[(a)+1]<<8) | M[a];
}

void setword(uint a, uint w) {
	M[(a)+1] = (w)>>8;
	M[a] = (w);
}

uint pop(void) {
	uint	v;

	v = word(Sp);
	Sp += 2;
	return v;
}

void push(uint v) {
	Sp -= 2;
	setword(Sp, v);
}

uint rpop(void) {
	uint	v;

	v = word(Bp);
	Bp += 2;
	return v;
}

void rpush(uint v) {
	Bp -= 2;
	setword(Bp, v);
}

void fini(void);

void lodsw(void) { Ax = word(Si); Si += 2; }

void next(void) { lodsw(); Bx = Ax; Dx = Bx + 2; }

void and(void) { push(pop() & pop()); next(); }
void branch(void) { Si = word(Si); next(); }
void bye(void) { fini(); next(); }
void c_fetch(void) { push(M[pop()]); next(); }
void c_store(void) { uint a; a = pop(); M[a] = pop(); next(); }

void cold(void) {
	int	i;

	Sp = 0xF400;
	Bp = 0xF780;
	Bdev = open("t3xforth.blk", O_RDWR);
	for (i=256; i<8192; i++)
		if (!memcmp(&M[i], "\005RESET", 6))
			break;
	if (i == 1280) quit("entry point not found", "RESET");
	Si = pfa(i);
	next();
}

void docons(void) { push(word(Dx)); next(); }

void dodo(void) {
	uint	a;

	a = pop();
	rpush(pop());
	rpush(a);
	next();
}

void does(void) {
	rpush(Si);
	Bx = Dx;
	Si = word(Bx);
	Dx += 2;
	push(Dx);
	next();
}

void docolon(void) { rpush(Si); Si = Dx; next(); }
void dolit(void) { lodsw(); push(Ax); next(); }
void dovar(void) { push(Dx); next(); }
void drop(void) { pop(); next(); }
void dup_(void) { push(word(Sp)); next(); }
void execute(void) { Dx = pop(); Bx = Dx-2; /* no next() */ }
void exit_(void) { Si = rpop(); next(); }
void fetch(void) { push(word(pop())); next(); }
void i_(void) { push(word(Bp)); next(); }

int extend(uint x) {
	if (x & 0x8000) return -(0x10000-x);
	return x;
}

void loop(void) {
	int	i;

	i = extend(rpop());
	if (i+1 < extend(word(Bp))) {
		rpush(i+1);
		Si = word(Si);
	}
	else {
		rpop();
		Si += 2;
	}
	next();
}

void lshift(void) { uint w; w = pop(); push(pop() << w); next(); }
void or(void) { push(pop() | pop()); next(); }
void over(void) { push(word(Sp+2)); next(); }

void plus_loop(void) {
	int	i, k;

	k = extend(pop());
	i = extend(rpop()) + k;
	if (((i - extend(word(Bp))) ^ k) < 0) {
		rpush(i);
		Si = word(Si);
	}
	else {
		rpop();
		Si += 2;
	}
	next();
}

void q_same(void) {
	uint	s, d, n;

	d = pop();
	s = pop();
	n = M[d] & 0x1f;
	if (n > 5) n = 5;
	if (M[s] == (M[d] & 0x1F) && memcmp(&M[s+1], &M[d+1], n) == 0)
		push(-1);
	else
		push(0);
	next();
}

void qbranch(void) {
	if (pop())
		Si += 2;
	else
		Si = word(Si);
	next();
}

void r_fetch(void) { push(word(Bp)); next(); }
void r_from(void) { push(rpop()); next(); }

void read_block(void) {
	if (lseek(Bdev, pop()*1024, SEEK_SET) < 0) {
		pop();
		push(0);
		next();
		return;
	}
	if (read(Bdev, &M[pop()], 1024) != 1024) {
		push(0);
		next();
		return;
	}
	push(-1);
	next();
}

void rp_fetch(void) { push(Bp); next(); }
void rp_store(void) { Bp = pop(); next(); }
void rshift(void) { uint w; w = pop(); push(pop() >> w); next(); }

void writes(char *s) { write(1, s, strlen(s)); }

void rx_question(void) {
	int	k;

	k = getch();
	if (k == ERR) {
		usleep(50000);
		push(0);
	}
	else {
		if (k == 10) k = 13;
		push(k);
		push(-1);
	}
	next();
}

void sp_fetch(void) { push(Sp); next(); }
void sp_store(void) { Sp = pop(); next(); }

void store(void) {
	uint a;

	a = pop();
	setword(a, pop());
	next();
}

void swap(void) {
	uint a, b;

	a = pop();
	b = pop();
	push(a);
	push(b);
	next();
}

void to_r(void) { rpush(pop()); next(); }
void tx_store(void) { char c; c = pop(); write(1, &c, 1); next(); }

void um_plus(void) {
	uint	a;

	a = pop() + pop();
	push(a & 0xFFFF);
	push(a >> 16);
	next();
}

void up_loop(void) {
	uint	i, k;

	k = pop();
	i = rpop();
	if (i+k < word(Bp)) {
		rpush(i+k);
		Si = word(Si);
	}
	else {
		rpop();
		Si += 2;
	}
	next();
}

void write_block(void) {
	if (lseek(Bdev, pop()*1024, SEEK_SET) < 0) {
		pop();
		push(0);
		next();
		return;
	}
	if (write(Bdev, &M[pop()], 1024) != 1024) {
		push(0);
		next();
		return;
	}
	push(-1);
	next();
}

void xor(void) { push(pop() ^ pop()); next(); }
void zero_less(void) { push(extend(pop()) < 0? -1: 0); next(); }

int	Break;

void setbreak(int dummy) {
	Break = 1;
}

void shutdown(void);
void init(void);

void prompt(void) {
	int	c;

	signal(SIGINT, setbreak);
	Break = 0;
	writes("\r\nTFVM: (b)reak (r)eboot (q)uit (c)ontinue: ");
	for (;;) {
		c = getch();
		if (c == ERR)
			usleep(50000);
		else
			break;
	}
	switch (c) {
	case 'b': writes("break\r\n"); cold(); break;
	case 'r': writes("reboot"); shutdown(); init(); cold(); break;
	case 'q': writes("quit\r\n"); fini(); break;
	default:  writes("continue\r\n"); break;
	}
}

void invalid(void) { writes("\r\nTFVM: invalid opcode"); prompt(); }

void cycle(void) {
	signal(SIGINT, setbreak);
	Break = 0;
	for (;;) {
		if (Break) prompt();
		Bx = word(Bx);
		if (Bx > Barrier) {
			invalid();
			continue;
		}
		Prim[Bx]();
	}
}

PRIM	Assignments[] = {
	{ "!    ", store	},
	{ "(+LOO", plus_loop	},
	{ "(/LOO", up_loop	},
	{ "(0BRA", qbranch	},
	{ "(BRAN", branch	},
	{ "(CONS", docons	},
	{ "(DO) ", dodo		},
	{ "(DOES", does		},
	{ "(LIT)", dolit	},
	{ "(COLO", docolon	},
	{ "(LOOP", loop		},
	{ "(VAR)", dovar	},
	{ "0<   ", zero_less	},
	{ ">R   ", to_r		},
	{ "?SAME", q_same	},
	{ "@    ", fetch	},
	{ "AND  ", and		},
	{ "BYE  ", bye		},
	{ "C!   ", c_store	},
	{ "C@   ", c_fetch	},
	{ "COLD ", cold		},
	{ "DROP ", drop		},
	{ "DUP  ", dup_		},
	{ "EXECU", execute	},
	{ "EXIT ", exit_	},
	{ "I    ", i_		},
	{ "LSHIF", lshift	},
	{ "OR   ", or		},
	{ "OVER ", over		},
	{ "R>   ", r_from	},
	{ "R@   ", r_fetch	},
	{ "READ-", read_block	},
	{ "RP!  ", rp_store	},
	{ "RP@  ", rp_fetch	},
	{ "RSHIF", rshift	},
	{ "RX?  ", rx_question	},
	{ "SP!  ", sp_store	},
	{ "SP@  ", sp_fetch	},
	{ "SWAP ", swap		},
	{ "TX!  ", tx_store	},
	{ "UM+  ", um_plus	},
	{ "WRITE", write_block	},
	{ "XOR  ", xor		},
	{  NULL,   NULL		}
};

void assign(uint a) {
	int	i;
	char	s[6];

	for (i=0; Assignments[i].name != NULL; i++) {
		if (memcmp(Assignments[i].name, &M[a+1], 5) == 0) {
			Prim[pfa(a)] = Assignments[i].fn;
			return;
		}
	}
	memcpy(s, &M[a+1], 5);
	s[5] = 0;
	quit("unknown primitive", s);
}

void findprims(void) {
	uint	i;

	for (i=256; i<1280; i++)
		if (!memcmp(&M[i], "\013WRITE", 6))
			break;
	if (i == 1280) quit("bad image file", NULL);
	Barrier = pfa(i);
	for (i=0; i<1280; i++) Prim[i] = invalid;
	i = nfa(Barrier);
	while (i != 0) {
		assign(i);
		i = word(lfa(i));
	}
}

#ifdef RAWMODE
 #define RAW() raw()
 #define COOKED() noraw()
#else
 #define RAW() cbreak()
 #define COOKED() nocbreak()
#endif

void init(void) {
	int	fd, k;

	memset(M, 0, 65536);
	fd = open("t3xforth.com", O_RDONLY);
	if (fd < 0) quit("image not found", NULL);
	k = read(fd, &M[256], 32768);
	close(fd);
	if (k < 7000 || k == 32768) quit("image read error", NULL);
	findprims();
	initscr();
	clear();
	refresh();
	RAW();
	noecho();
	nodelay(stdscr, 1);
}

void shutdown(void) {
	close(Bdev);
	COOKED();
	echo();
	nodelay(stdscr, 0);
	endwin();
}

void fini(void) {
	move(24, 0);
	shutdown();
	exit(0);
}

int main(void) {
	init();
	cold();
	cycle();
	fini();
	return 0;
}
